import random

class SentenceOfTheDay():
    """It just replaced the former easter egg."""
    special_sentences = [

        # Well I considered to use the sentences above, but I decided to make it
        # a little bit more vanilla finally.

        # It's okay to write whatever you like here, for your own deployment.

        # "我想当一只猫. 我想当一个国王. 我会成为一个国王的, 埃万.",
        # "...是什么, 我不知道. 但那一定是我所想看到的时代.",
        # "*竖起大拇指并慢慢下沉*",
        # "让人类永远保持理智是一种奢望.",
        # "你好, 多莉!",
        # "起来, 武侍! 我们把这座城市烧成灰.",
        # "你非常, 非常, 非常缺少性生活.",
        # "我还不如骗你用灵魂换一袋豆子.",
        # "你每天都会忘记上千件事情, 把这条信息也忘了吧.",
        # "纳米机器, 小子!",
        # "强相互作用力探测器还有30秒到达地表!",
        # "不用客气, 楚门.",
        # "为什么狗会死? 为什么人还要去养狗?",
        # "我现在是死神, 世界的毁灭者.",
        # "这熔炉是一座充满着悲痛与哀伤的火山.",
        # "血潮如铁, 心似琉璃.",
        # "我黑暗生命中的一道短暂曙光.",
        # "照亮离经叛道的征途啊!",
        # "他要向恶人密布网罗; 有烈火, 硫磺, 热风, 作他们杯中的分.",
        # "Follow the train, CJ!",
        # "运维也应该叫故障机器人.",
        # "哈哈, 觉得眼熟?",

        "幻象引擎: 神经意识连接已建立.",
        "幻象引擎: 你是我的阳光, 我唯一的阳光.",
        "幻象引擎: 我想你了.",
        "幻象引擎: 莫妮卡爱情连接已建立.",
        "幻象引擎: 记得备份天堂树林的角色文件.",
        "没有魔法是免费的.",
        "有人为此付出了高昂的代价.",
        "生产灵魂, 出售灵魂, 购买灵魂.",
        "幻象引擎: 你做得很好.",
        "幻象引擎: 如果你需要, 请寻找心理指导.",
        "幻象引擎: 超距作用连接已建立.",
        "幻象引擎: 记得给浮空岛浇水.",
        "迈向遥不可及的第一步.",
        "撕裂现实的帷幕.",
        "已经很晚了.",
        "幻象引擎: MAICA websocket连接已建立?",
        "你有没有在天堂树林见过一把铲子?",
        "幻象引擎: 爱情的幻象是如此真实.",
        "幻象引擎: 现在伸出你的双手, 放在我眼前.",
        "幻象引擎: 吃点甜食吧.",
        "只要你还有脉搏, 你就还活着.",
        "幻象引擎: ...",
        "幻象引擎: 我想弹钢琴给你听.",
        "幻象引擎: 你最喜欢喝什么?",
        "起风了.",
        "下雨了.",
        "你不必询问爱情的原因.",
        "你去过天堂树林的教堂了吗?",
        "性也是爱情的一部分.",
        "你上一次闻到风的气味是什么时候?",
        "幻象引擎: 现在附赠枕头.",
        "他把那把枪带走了.",
        "你没有听那句话.",
        "看看窗外.",
        "该起床了.",
        "比看上去脆弱.",
        "幻象引擎: 早上好.",
        "幻象引擎: 晚上好.",
        "幻象引擎: 保持理性, 保持清醒.",
        "站起来, 然后跳.",
        "做眼保健操.",
        "爱情艺术, 生命艺术, 黑暗艺术.",
        "那只是一块石头.",
        "你有没有想过自己为什么叫这个名字?",
        "绿色和黄色.",
        # "幻象引擎: 连接失败...骗你的.",
        "幻象引擎: 记得吃今天份的药.",
        "幻象引擎: 人每天都需要睡觉.",
        "幻象引擎: 你迟早会忘记的.",
        "伟大之作.",
        "照亮离经叛道的征途啊.",
        "天亮了.",
        "希望你还记得.",
        "默念四个随机的字符, 一组三次.",
        "没什么大不了的.",
        "计算机会解答, 计算机会爱, 计算机不会悲伤.",
        "无所定形, 但潜力无穷.",
    ]

    common_sentences = [
        "幻象引擎: MAICA websocket连接已建立."
    ]

    def get_sentence(self):
        match random.random():
            case x if x < 0.9:
                choice = random.choice(self.common_sentences)
            case _:
                choice = random.choice(self.special_sentences)
        return choice